#  Copyright (c) 2025. #  OCX Consortium https://3docx.org. See the LICENSE
from ocx_common.utilities.ocx_xml import OcxXml
from tests.conftest import MODEL_FOLDER, NAMESPACE, SCHEMA_VERSION, TEST_MODEL


class TestOcxXml:
    def test_get_version(self, shared_datadir):
        model = shared_datadir / MODEL_FOLDER / TEST_MODEL
        assert OcxXml.get_version(str(model.resolve())) == SCHEMA_VERSION

    def test_get_ocx_namespace(self, shared_datadir):
        model = shared_datadir / MODEL_FOLDER / TEST_MODEL
        assert OcxXml.get_ocx_namespace(str(model.resolve())) == NAMESPACE

    def test_get_all_namespaces(self, shared_datadir, data_regression):
        model = shared_datadir / MODEL_FOLDER / TEST_MODEL
        result = OcxXml.get_all_namespaces(str(model.resolve()))
        data_regression.check(result)

    def test_has_ocx_namespace(self, shared_datadir):
        model = shared_datadir / MODEL_FOLDER / TEST_MODEL
        assert OcxXml.has_ocx_namespace(str(model.resolve())) == True

    def test_has_unitsml_namespace(self, shared_datadir):
        model = shared_datadir / MODEL_FOLDER / TEST_MODEL
        assert OcxXml.has_unitsml_namespace(str(model.resolve())) == True



#  Copyright (c) 2023. OCX Consortium https://3docx.org. See the LICENSE
"""Interfaces module."""

# System imports
import threading
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterator, Optional
from uuid import uuid4
from enum import Enum

from requests.models import Response


class _Singleton:
    """A threadsafe singleton class.
    https://medium.com/gitconnected/design-patterns-in-python-singleton-pattern-f76dc26281f8
    """

    __instance = None
    __lock = threading.Lock()

    def __init__(self):
        """
        Initializes a new instance with a randomly generated UUID value.

        Args:
            self: The instance itself.

        Returns:
            None
        """
        self.value = uuid4()

    @classmethod
    def get_instance(cls):
        """
        Returns the instance of the class. If the instance does not exist, it creates one.

        Returns:
            The instance of the class.

        Examples:
            instance = ClassName.get_instance()
        """
        with cls.__lock:
            if not cls.__instance:
                cls.__instance = cls()
        return cls.__instance


class IModuleDeclaration(ABC):
    """Abstract module import declaration Interface"""

    @abstractmethod
    def get_declaration(self) -> str:
        """Abstract Method: Return the module declaration string."""
        pass


class IObserver(ABC):
    """The observer interface"""

    @abstractmethod
    def update(self, event: str, payload: Dict):
        """Interface update method"""


class IObservable(ABC):
    """Interface. The observable object."""

    @abstractmethod
    def subscribe(self, observer: IObserver):
        """subscription"""

    @abstractmethod
    def unsubscribe(self, observer: IObserver):
        """unsubscribe"""

    @abstractmethod
    def update(self, event: Enum, message: Dict):
        """
        update method.

        Args:
            event: The event type
            message: The event message
        """


class IParser(ABC):
    """Abstract IParser interface."""

    @abstractmethod
    def parse(self, model: str) -> Optional[Any]:
        """
        Abstract method for parsing a data model,

        Args:
            model: the data model source

        Returns:
            the root dataclass of the parsed data model.
        """
        pass

    @abstractmethod
    def iterator(self, model) -> Iterator:
        """
        Abstract method for iterating a data model.

        Args:
            model: the data model to iterate on.
        Returns:
             An iterator
        """
        pass


class ISerializer(ABC):
    """OcxSerializer interface"""

    def __init__(self, model: Any):
        self.model = model

    @abstractmethod
    def serialize_to_file(self, to_file: str) -> bool:
        """Abstract XML serialize to file method"""
        pass

    @abstractmethod
    def serialize_to_string(self) -> str:
        """Abstract XML serialize to string method"""
        pass


class IRestClient(ABC):
    """Abstract IRestClient interface for REST server clients."""

    def __init__(self, headers: Dict, timeout: int = 30):
        """

        Args:
            headers: The request headers
            timeout: The request timeout

        Parameters:
            base_url: base url
            timeout: Timeout in ms
        """
        self._timeout: int = timeout
        self._headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        if headers is not None:
            self._headers = headers

    @abstractmethod
    def get(self, url: str, headers: Dict) -> Response:
        """Abstract get method"""

    @abstractmethod
    def post(self, url: str, payload: Dict, headers: Dict) -> Response:
        """Abstract post method"""

    @abstractmethod
    def delete(self, url: str, payload: Dict, headers: Dict) -> Response:
        """Abstract delete  method"""

    @abstractmethod
    def put(self, url: str, payload: Dict, headers: Dict) -> Response:
        """Abstract put  method"""

    @abstractmethod
    def patch(self, url: str, payload: Dict, headers: Dict) -> Response:
        """Abstract patch  method"""

    @abstractmethod
    def set_headers(self, headers: Dict):
        """Abstract set headers method"""



#  Copyright (c) 2024.  OCX Consortium https://3docx.org. See the LICENSE
"""Reusable decorators"""

# System imports
import time
from functools import wraps
from typing import Dict

# Third party imports
from loguru import logger
from icecream import ic


def timer(func):
    """@timer decorator for measuring elapsed time of a function call."""

    @wraps(func)
    def inner_func(*args, **kwargs):
        # start the timer
        start_time = time.time()
        # call the decorated function
        result = func(*args, **kwargs)
        # remeasure the time
        end_time = time.time()
        # compute the elapsed time and print it
        execution_time = end_time - start_time
        print(f"Execution time: {execution_time} seconds")
        # return the result of the decorated function execution
        return result

    # return reference to the wrapper function
    return inner_func


def exception_handler(exception_to_raise: BaseException):
    """@exception_handler decorator"""

    def outer_func(func):
        @wraps(func)
        def inner_func(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                # Handle the exception
                print(f"An exception occurred: {str(e)}")
                logger.exception(f"An exception occurred: {str(e)}")
                # Optionally, perform additional error handling or logging
                # Reraise the exception if needed
                raise exception_to_raise from e

        return inner_func

    return outer_func


def debugger(func):
    """@debug decorator using ice cream"""

    def wrapper(*args, **kwargs):
        # print the function name and arguments
        for arg in args:
            ic(arg)
        # call the function
        result = func(*args, **kwargs)
        # print the results
        ic(result)
        logger.debug(f"{func.__name__} returned: {result}")
        return result

    return wrapper


def memoize(func):
    """@memoize decorator"""
    cache: Dict = {}

    def wrapper(*args, **kwargs):
        if args in cache:
            return cache[args]
        else:
            result = func(*args)
            cache[args] = result
            return result

    return wrapper



#  Copyright (c) 2024. #  OCX Consortium https://3docx.org. See the LICENSE

# Third party imports
from lxml import etree

from ocx_common.parser.parsers import OcxParser
from ocx_common.utilities.ocx_xml import OcxXml
from ocx_common.x_path.x_path import OcxPathBuilder
from ocx_common.x_path.xelement import LxmlElement

# Project imports
from .conftest import MODEL_FOLDER, NAMESPACE, SCHEMA_VERSION, TEST_MODEL


class TestXPathBuilder:
    def test_select_all_named_nodes(self, shared_datadir):
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxParser(str(file))
        root = parser.get_root()
        search = etree.XPath(
            path=OcxPathBuilder.select_all_named_nodes("RefPlane", namespace="ocx"),
            namespaces=OcxXml.get_all_namespaces(str(file)),
        )
        result = search(root)
        assert len(result) == 13

    def test_select_all_named_unitsml_nodes(self, shared_datadir):
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxParser(str(file))
        root = parser.get_root()
        print(root.nsmap)
        search = etree.XPath(
            path=OcxPathBuilder.select_all_named_nodes("Unit", namespace="unitsml"),
            namespaces=OcxXml.get_all_namespaces(str(file)),
        )
        result = search(root)
        assert len(result) == 17

    def test_select_current_node_is_root(self, shared_datadir):
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxParser(str(file))
        root = parser.get_root()
        search = etree.XPath(
            path=OcxPathBuilder.select_current_node(),
            namespaces=OcxXml.get_all_namespaces(str(file)),
        )
        result = search(root)
        node = result[0]
        assert (
            node.get("schemaVersion") == SCHEMA_VERSION
            and node.tag == f"{LxmlElement.namespaces_decorate(NAMESPACE)}ocxXML"
        )

    def test_select_current_node_is_vessel(self, shared_datadir):
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxParser(str(file))
        root = parser.get_root()
        vessel = LxmlElement.find_child_with_name(element=root, child_name="Vessel")
        search = etree.XPath(
            path=OcxPathBuilder.select_current_node(),
            namespaces=OcxXml.get_all_namespaces(str(file)),
        )
        result = search(vessel)
        node = result[0]
        assert (
            node.get("name") == "OCX-MODEL1/A"
            and node.tag == f"{LxmlElement.namespaces_decorate(NAMESPACE)}Vessel"
        )

    def test_select_parent_node_is_vessel(self, shared_datadir):
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxParser(str(file))
        root = parser.get_root()
        panel = LxmlElement.find_child_with_name(root, "Panel")
        search = etree.XPath(
            path=OcxPathBuilder.select_parent_node(),
            namespaces=OcxXml.get_all_namespaces(str(file)),
        )
        result = search(panel)
        parent = result[0]
        assert (
            parent.get("id") == "nplcid1"
            and parent.tag == f"{LxmlElement.namespaces_decorate(NAMESPACE)}Vessel"
        )

    def test_select_named_parent_node_is_coordinate_system(self, shared_datadir):
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxParser(str(file))
        root = parser.get_root()
        x_ref = LxmlElement.find_child_with_name(root, "XRefPlanes")
        search = etree.XPath(
            path=OcxPathBuilder.select_named_parent_node(parent="CoordinateSystem"),
            namespaces=OcxXml.get_all_namespaces(str(file)),
        )
        result = search(x_ref)
        parent = result[0]
        assert (
            parent.get("id") == "nplcid7"
            and parent.tag
            == f"{LxmlElement.namespaces_decorate(NAMESPACE)}CoordinateSystem"
        )

    def test_select_named_node_xrefplanes(self, shared_datadir):
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxParser(str(file))
        root = parser.get_root()
        search = etree.XPath(
            path=OcxPathBuilder.select_named_nodes("XRefPlanes"),
            namespaces=OcxXml.get_all_namespaces(str(file)),
        )
        result = search(root)
        assert len(result) == 1

    def test_select_all_nodes_with_attribute_name_guidref(self, shared_datadir):
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxParser(str(file))
        root = parser.get_root()
        search = etree.XPath(
            path=OcxPathBuilder.select_any_nodes_with_global_attribute_name("GUIDRef"),
            namespaces=OcxXml.get_all_namespaces(str(file)),
        )
        result = search(root)
        assert len(result) == 35

    def test_select_named_nodes_with_attribute_name_panel(self, shared_datadir):
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxParser(str(file))
        root = parser.get_root()
        search = etree.XPath(
            path=OcxPathBuilder.select_named_nodes_with_global_attribute_name(
                node_name="Panel", attribute_name="GUIDRef"
            ),
            namespaces=OcxXml.get_all_namespaces(str(file)),
        )
        result = search(root)
        assert (
            len(result) == 1
            and result[0].tag == f"{LxmlElement.namespaces_decorate(NAMESPACE)}Panel"
        )

    def select_named_nodes_with_attribute_value_of_guidref(self, shared_datadir):
        guidref = "6e17c799-8e76-416d-ba75-ca0565dd3c42"
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxParser(str(file))
        root = parser.get_root()
        search = etree.XPath(
            path=OcxPathBuilder.select_named_nodes_with_global_attribute_value(
                node_name="Panel", attribute_name="GUIDRef", attribute_value=guidref
            ),
            namespaces=OcxXml.get_all_namespaces(str(file)),
        )
        result = search(root)
        assert len(result) == 1 and result[0].get("GUIDRef") == guidref

    def select_any_nodes_with_attribute_value(self, shared_datadir):
        guidref = "6e17c799-8e76-416d-ba75-ca0565dd3c42"
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxParser(str(file))
        root = parser.get_root()
        search = etree.XPath(
            path=OcxPathBuilder.select_any_nodes_with_attribute_value(
                attribute_name="GUIDRef", attribute_value=guidref
            ),
            namespaces=OcxXml.get_all_namespaces(str(file)),
        )
        result = search(root)
        assert len(result) == 1 and result[0].get("GUIDRef") == guidref

    def test_select_all_nodes_with_attribute_name(self, shared_datadir):
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxParser(str(file))
        root = parser.get_root()
        search = etree.XPath(
            path=OcxPathBuilder.select_named_nodes_with_local_attribute_name(
                node_name="EnumeratedRootUnit",
                attribute_name="prefix",
                namespace="unitsml",
            ),
            namespaces=OcxXml.get_all_namespaces(str(file)),
        )
        result = search(root)

        assert len(result) == 9



#  Copyright (c) 2024. #  OCX Consortium https://3docx.org. See the LICENSE


# Project imports

from ocx_common.parser.parsers import OcxParser
from ocx_common.x_path.x_path import OcxGuidRef, OcxPath

from .conftest import MODEL_FOLDER, TEST_MODEL


def test_find_vessel(shared_datadir):
    file = shared_datadir / MODEL_FOLDER /TEST_MODEL
    parser = OcxParser(str(file))
    root = parser.get_root()
    ocx_path = OcxPath(root, namespaces=root.nsmap)
    node = ocx_path.get_all_named_ocx_elements(name="Vessel", namespace="ocx")
    assert len(node) == 1


def test_find_all_guids(shared_datadir):
    file = shared_datadir / MODEL_FOLDER /TEST_MODEL
    parser = OcxParser(str(file))
    root = parser.get_root()
    guids = OcxGuidRef(root, namespaces=root.nsmap).get_all_guids()
    assert len(guids) == 25


def test_find_child_guids(shared_datadir):
    file = shared_datadir / MODEL_FOLDER / TEST_MODEL
    parser = OcxParser(str(file))
    root = parser.get_root()

    guids = OcxGuidRef(root, namespaces=root.nsmap).get_child_guids("CoordinateSystem")
    assert len(guids) == 25


def test_get_all_named_nodes(shared_datadir):
    file = shared_datadir / MODEL_FOLDER / TEST_MODEL
    parser = OcxParser(str(file))
    root = parser.get_root()
    nodes = OcxPath(root, namespaces=root.nsmap).get_all_named_ocx_elements(
        "CoordinateSystem"
    )
    assert len(nodes) == 1



# -*- coding: utf-8 -*-

#  Copyright (c) 2023. OCX Consortium https://3docx.org. See the LICENSE

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# This file is execfile()d with the current directory set to its
# containing dir.

from __future__ import unicode_literals

import os
import sys
import datetime

sys.path.insert(0, os.path.abspath(".."))

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx_autodoc_typehints",
]
# Add the following lines to include docstrings from class __init__ methods
autoclass_content = 'both'
source_suffix = ".rst"
master_doc = "index"
project = "ocx-common"
year = "2025"
author = "Ole Christian Astrup"
organisation= '3Docx.org'
copyright = f"{datetime.datetime.now().year}, {organisation}"
version = "2.1.3"
release = version
pygments_style = "trac"
templates_path = ["_templates"]
extlinks = {
    "issue": ("OCXStandard/ocx-common/issues/%s", "#"),
    "pr": ("OCXStandard/ocx-common/pull/%s", "PR #"),
}
# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

if not on_rtd:  # only set the theme if we're building docs locally
    html_theme = "sphinx_rtd_theme"

html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
html_split_index = False
html_sidebars = {
    "**": ["searchbox.html", "globaltoc.html", "sourcelink.html"],
}
html_short_title = "%s-%s" % (project, version)
napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False



#  Copyright (c) 2023-2025. OCX Consortium https://3docx.org. See the LICENSE

from pathlib import Path
import pytest
from ocx_common.utilities.downloader import SchemaDownloader
from tests.conftest import NAMESPACE

@pytest.mark.skip(reason="Remove this functionality")
def test_download_from_url(shared_datadir):
    """Test download from external source."""

    downloader = SchemaDownloader(shared_datadir)
    downloader.wget(NAMESPACE)
    files = list(shared_datadir.glob("*.xsd"))
    assert len(files) == 3


@pytest.mark.skip(reason="Remove this functionality")
def test_download_from_file(shared_datadir: Path):
    """Test download from a local file."""
    # Temp folder
    temp = shared_datadir / "temp"
    downloader = SchemaDownloader(temp)
    ocx_schema = shared_datadir / "schemas/OCX_Schema.xsd"
    downloader.wget(str(ocx_schema.resolve()))
    files = list(temp.glob("*.xsd"))
    assert len(files) == 3



#  Copyright (c) 2024. OCX Consortium https://3docx.org. See the LICENSE

import os
import sys

from icecream import ic
from loguru import logger

# Disable or enable debug/logging
ic.enable()
ic.configureOutput(includeContext=True, contextAbsPath=False)
logger.disable("ocx_common")

# To make sure that the tests import the modules this has to come before the import statements
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../")))



SCHEMA_VERSION = "3.0.0"
NAMESPACE = "https://3docx.org/fileadmin//ocx_schema//V300//OCX_Schema.xsd"
MODEL1 = "NAPA-OCX_M1.3docx"

MOCK_URL = "http://localhost:8080/rest/api"
MODEL_FOLDER = "models"
SCHEMA_FOLDER = "schemas"
TEST_MODEL = MODEL1



#  Copyright (c) 2025. #  OCX Consortium https://3docx.org. See the LICENSE


import os
import urllib.parse
from pathlib import Path

import requests

# Third party imports
from loguru import logger

from ocx_common.decorators.decorators import exception_handler

# Project imports
from ocx_common.utilities.utilities import (
    is_local_file_uri,
    is_valid_absolute_windows_path,
    is_valid_unix_file_path,
    is_windows_drive_letter,
)


class URIError(BaseException):
    """URIValidator errors."""


class URIValidator:
    """
    URI and OS source validation.
    Check both correctness and existence of the source.
    """

    def __init__(self, uri: str):
        self.uri = uri

    @exception_handler(BaseException)
    def is_valid(self, check_source: bool = False) -> bool:
        """Validates whether the input is a correct URI or a valid file path.

        Args:
            check_source: If True, check if the source of the URI exists.
        Raises:
            URIError if the source does not exist and check_source is True
        """
        # The uri is an OS file path
        if self._is_valid_file_path(self.uri):
            if check_source:
                return self._file_exists(self.uri)
            else:
                return True
        else:
            try:
                parsed = urllib.parse.urlparse(self.uri)
                if parsed.scheme == "http" or parsed.scheme == "https":
                    return self._is_valid_uri(parsed) and self._uri_exists(check_source)
                elif parsed.scheme == "file":
                    if parsed.path == "":
                        logger.error(f"Missing valid absolute path in {self.uri}")
                        return False
                    elif is_local_file_uri(self.uri):
                        return is_valid_absolute_windows_path(
                            parsed.path
                        ) and self._is_valid_file_path(parsed.path)
                    else:
                        return False
                else:
                    e = f"Existence of the scheme {parsed.scheme} is not supported"
                    logger.error(e)
                    if check_source:
                        raise URIError(e)
                    else:
                        return self._is_valid_uri(parsed)

            except (ValueError, URIError) as e:
                logger.error(e)
                raise URIError from e

    def _is_valid_file_path(self, file_path: str) -> bool:
        """Checks if the given path is a valid file path."""
        # on Windows
        if os.name == "nt":
            return is_valid_absolute_windows_path(file_path)
        # On UNIX like os
        else:
            return is_valid_unix_file_path(file_path)

    def _is_valid_uri(self, parsed) -> bool:
        """Validates whether the given URI has a correct structure."""
        return bool(parsed.scheme and parsed.netloc)

    def _file_exists(self, path: str) -> bool:
        path_to_file = Path(path)
        return path_to_file.exists()

    def _uri_exists(self, check_source: bool) -> bool:
        if not check_source:
            return True
        response = requests.head(self.uri)
        if response.status_code == 200:
            logger.info(f"The {self.uri} is accessible")
            return True
        elif response.status_code == 404:
            msg = f"The URI {self.uri} does not exist. Server response: {response.text}"
            logger.error(msg)
            raise URIError(msg)
        else:
            e = f"(The request returned {response.text}"
            raise URIError(e)

    def is_local_file(self) -> bool:
        if is_valid_absolute_windows_path(self.uri):
            return True
        parsed = urllib.parse.urlparse(self.uri)
        if parsed.scheme == "file":
            if is_windows_drive_letter(parsed.netloc):
                return True
            else:
                return False
        elif parsed.scheme == "":
            return is_valid_absolute_windows_path(self.uri)
        else:
            return False



#  Copyright (c) 2023. OCX Consortium https://3docx.org. See the LICENSE
"""Dynamically load a python module."""

# system imports
import importlib
import sys
from abc import ABC
from types import ModuleType
from typing import Any, List
import importlib.util

# 3rd party imports
from loguru import logger

# Project imports
from ocx_common.interfaces.interfaces import IModuleDeclaration


class DynamicLoaderError(BaseException):
    """Dynamic import errors."""


class ModuleDeclaration(IModuleDeclaration, ABC):
    """General module declaration

    Args:
        package:the package name
        sub_module: The submodule-name
        name: The method name
    """

    def __init__(self, package: str, sub_module: str, name: str):
        self._module = sub_module
        self._package = package
        self._method = name

    def get_declaration(self) -> str:
        """Return the module import declaration."""
        return f"{self._package}.{self._module}.{self._method}"


class DeclarationOfOcxImport(IModuleDeclaration):
    """Declaration of the ocx module."""

    def __init__(self, name: str, version: str):
        self._name = name
        self._version = version

    def get_declaration(self) -> str:
        """Return the module import declaration."""
        ocx_pkg = f"{self._name}_{self._version.replace('.', '')}"
        return f"{self._name}.{ocx_pkg}.{ocx_pkg}"

    def get_version(self) -> str:
        """Return the OCX module version."""
        return self._version

    def get_name(self) -> str:
        """Return the declared module name."""
        return self._name


class DynamicLoader:
    """Dynamically loads modules, classes of functions from a module declaration."""

    @classmethod
    def _load(cls, declaration: str) -> Any:
        """Internal Method: Import the object from the declaration.

        Args:
            declaration: The module declaration string.
        Returns:
            Return the loaded object, None if failed.
        """
        obj_type = None
        if (spec := importlib.util.find_spec(declaration)) is not None:
            obj_type = importlib.util.module_from_spec(spec)
            sys.modules[declaration] = obj_type
            spec.loader.exec_module(obj_type)
            # logger.debug(f"Loaded object {declaration!r} from location {spec.origin!r}")
        else:
            logger.error(f"No object {declaration!r}")
        return obj_type

    @classmethod
    def import_module(cls, module_declaration: IModuleDeclaration) -> ModuleType:
        """

        Args:
            module_declaration: The declaration of the python module to load

        Returns:
            Return the loaded module, None if failed.
        """
        module_to_load = module_declaration.get_declaration()
        return cls._load(module_to_load)

    @classmethod
    def import_class(
        cls, module_declaration: IModuleDeclaration, class_name: str
    ) -> Any:
        """
        The module import declaration.

        Args:
            class_name: The class name to load form the declared module
            module_declaration: The declaration of the python module to be loaded

        Returns:
            Return the loaded class, None if failed.

        """

        obj = None
        module_to_load = module_declaration.get_declaration()
        module = cls._load(module_to_load)
        try:
            obj = getattr(module, class_name)
            # logger.debug(f"Loaded class {class_name!r}"
            return obj
        except AttributeError as e:
            raise DynamicLoaderError(
                f"No class with name {class_name!r} in module {module_to_load!r}"
            ) from e

    @classmethod
    def get_all_class_names(cls, module_name: str, version: str) -> List:
        """Return all class names in the module by the ``__all__`` variable.

        Args:
            module_name: The module name
            version: The module version

        Returns:
            The list of available module class names.


        Example:
            >>> from ocx_common.loader.loader import DeclarationOfOcxImport , DynamicLoader
            >>> class_name = "Vessel"
            >>> declaration = DeclarationOfOcxImport("ocx", '3.0.1')
            >>> data_class = DynamicLoader.import_class(declaration, class_name)()
            >>> data_class.__doc__
            Vessel asset subject to Classification.

        """
        all_names = []
        ocx_pkg = f"{module_name}_{version.replace('.', '')}"
        ocx_module = f"{module_name}.{ocx_pkg}"
        if (spec := importlib.util.find_spec(ocx_module)) is not None:
            module = importlib.util.module_from_spec(spec)
            sys.modules[ocx_module] = module
            spec.loader.exec_module(module)
            logger.debug(f"Found module {ocx_module!r} in location {spec.origin}")
            all_names = module.__all__
        else:
            logger.error(f"No module with name {module_name!r} and version {version!r}")
        return all_names



#  Copyright (c) 2025. #  OCX Consortium https://3docx.org. See the LICENSE
"""Mock server for client requests."""

import pytest
import responses

from tests.conftest import MOCK_URL


@pytest.fixture
def mock_server():
    """Mock server for HTTP requests."""
    endpoints = [
        {
            "method": responses.GET,
            "url": f"{MOCK_URL}/get_200",
            "json": {"id": 1, "name": "John Doe"},
            "status": 200,
        },
        {
            "method": responses.POST,
            "url": f"{MOCK_URL}/post_201",
            "json": {"id": 1, "name": "John Doe"},
            "status": 201,
        },
        {
            "method": responses.GET,
            "url": f"{MOCK_URL}/get_400",
            "json": {"error": "Bad Request"},
            "status": 400,
        },
    ]

    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        for endpoint in endpoints:
            rsps.add(
                method=endpoint["method"],
                url=endpoint["url"],
                json=endpoint["json"],
                status=endpoint["status"],
            )
        yield rsps



#  Copyright (c) 2024. #  OCX Consortium https://3docx.org. See the LICENSE

# Project imports
from ocx_common.parser.parsers import OcxNotifyParser, OcxParser

from .conftest import MODEL_FOLDER, TEST_MODEL


class TestNotifyParser:
    def test_parse(self, shared_datadir):
        file = shared_datadir / MODEL_FOLDER / TEST_MODEL
        parser = OcxNotifyParser()
        root = parser.parse(str(file))
        name = root.header.name
        assert name == "OCX-MODEL1/A"


def test_parser(shared_datadir):
    file = shared_datadir / MODEL_FOLDER / TEST_MODEL
    parser = OcxParser(str(file))
    root = parser.get_root()
    assert (
        root.tag
        == "{https://3docx.org/fileadmin//ocx_schema//V300//OCX_Schema.xsd}ocxXML"
    )


# def test_parser_invalid_source(shared_datadir):   # ToDO: Add edge test case when source does not exist
#     file = shared_datadir / "not_exist.3docx"
#     try:
#         OcxParser(str(file))
#         assert False
#     except OcxParserError:
#         assert True



#  Copyright (c) 2023-2024. #  OCX Consortium https://3docx.org. See the LICENSE
"""Shared utility classes and functions"""

# System imports
import errno
import os
import re
import sys
import urllib.parse
from collections import defaultdict
from itertools import groupby
from pathlib import Path
from typing import Any, Dict, Generator, List

# Third party imports
# Project imports


def is_substring_in_list(substring, string_list):
    """

    Args:
        substring: The search string
        string_list: List of strings

    Returns:
        True if the substring is found, False otherwise.
    """
    return any(substring in string for string in string_list)


def all_equal(iterable) -> bool:
    """
    Verify that all items in a list are equal
    Args:
        iterable:

    Returns:
        True if all are equal, False otherwise.
    """
    g = groupby(iterable)
    return next(g, True) and not next(g, False)


def resource_path(relative_path):
    """Get absolute path to resource, works for dev and for PyInstaller"""
    base_path = getattr(sys, "_MEIPASS", os.path.abspath(".."))
    return os.path.join(base_path, relative_path)


def parent_directory(file: str) -> str:
    """The full path to the folder containing the ``file``

    Args:
        file: The name of an existing file
    """
    return os.path.realpath(os.path.join(os.path.dirname(file), ""))


def nested_dict():
    """
    A recursive function that creates a default dictionary where each value is
    another default dictionary.
    """
    return defaultdict(nested_dict)


def get_key_from_value(my_dict, value) -> Any:
    """
        Return the key associated with the value
    Args:
        my_dict: The dictionary of key, value pairs
        value: The value to search for

    Returns:
        The key if found
    Raises:
        ValueError if the value does not exist in the dictionary.
    """
    for key, val in my_dict.items():
        if val == value:
            return key
    raise ValueError(
        f"The value {value} does not exist in the dictionary"
    )  # raise a ValueError if the value is not found


def default_to_regular(d) -> Dict:
    """
    Converts defaultdict of defaultdict to dict of dicts.

    Args:
        d: The dict to be converted

    """
    if isinstance(d, defaultdict):
        d = {k: default_to_regular(v) for k, v in d.items()}
    return d


def default_to_grid(d) -> Dict:
    """
    Converts defaultdicts to a data grid with unique row ids.

    Args:
        d: The dict to be converted

    """
    if isinstance(d, defaultdict):
        print(d.items())
        d = {k: default_to_regular(v) for k, v in d.items()}
    return d


def list_files_in_directory(directory: str, filter: str) -> List:
    """
    Lists files in a directory based on a specified filter using glob pattern matching.

    Args:
        directory (str): The path to the directory.
        filter (str): The filter pattern to apply when listing files.

    Returns:
        List: A list of file paths that match the filter criteria.
    Raises:
        AssertionError: If the directory does not exist.
    """
    dir_path = Path(directory)
    if not dir_path.is_dir():
        raise AssertionError(errno.EEXIST)
    return [file.name for file in dir_path.glob(filter) if file.is_file()]


def camel_case_split(str) -> List:
    """Split camel case string to individual strings."""
    return re.findall(r"[A-Z](?:[a-z]+|[A-Z]*(?=[A-Z]|$))", str)


def dromedary_case_split(str) -> List:
    """Split camel case string to individual strings."""
    return re.findall(r"[A-Z]?[a-z]+|[A-Z]+(?=[A-Z]|$)", str)


def get_file_path(file_name):
    """Get the correct file path also when called within a one-file executable."""
    base_path = sys._MEIPASS if hasattr(sys, "_MEIPASS") else os.path.abspath("..")
    return os.path.join(base_path, file_name)


def is_valid_absolute_windows_path(path: str) -> bool:
    """
    Returns True if the path is a valid absolute Windows path.

    Args:
        path (str): The path to validate.

    Returns:
        bool: True if valid, False otherwise.
    """
    if not path:
        return False

    # Normalize slashes (convert forward slashes to backslashes)
    path = path.replace("/", "\\")

    # Windows MAX_PATH limit (excluding long path support)
    if len(path) > 260:
        return False

    # Ensure the path starts with a valid drive letter or UNC path
    drive_letter_pattern = re.compile(r"^[a-zA-Z]:\\")
    unc_pattern = re.compile(r"^\\\\[^\\/:*?\"<>|\r\n]+\\[^\\/:*?\"<>|\r\n]+")

    if not (drive_letter_pattern.match(path) or unc_pattern.match(path)):
        return False

    # Ensure `os.path.isabs()` confirms it as an absolute path
    if not os.path.isabs(path):
        return False

    return True


def is_windows_drive_letter(scheme: str) -> bool:
    """

    Args:
        scheme: The urlparse scheme

    Returns:
        True if the scheme is a valid Windows drive letter
    """
    return bool(re.fullmatch(r"[A-Za-z]:", scheme))


def is_valid_unix_file_path(path: str) -> bool:
    """
        Return True if the path is a valid UNIX path
    Args:
        path: path to validate

    Returns: True if validated, False otherwise.

    """
    unix_pattern = re.compile(
        r"^(\/|\.{1,2}\/|~\/|\.)(?:[a-zA-Z0-9._-]+\/)*[a-zA-Z0-9._-]+\/?$"
    )
    return bool(unix_pattern.fullmatch(path))


def is_valid_file_path(path: str) -> bool:
    """

    Args:
        path:

    Returns:

    """
    if os.name == "nt":
        return is_valid_absolute_windows_path(path)
    else:
        return is_valid_unix_file_path(path)


def is_local_file_uri(uri: str) -> bool:
    """Return True if the file uri is a local file"""
    parsed = urllib.parse.urlparse(uri)
    if parsed.scheme == "file" and is_windows_drive_letter(parsed.netloc):
        file_path = f"{parsed.netloc}{parsed.path}"
        return is_valid_file_path(file_path)
    else:
        return False


def file_uri_to_path(uri: str) -> str:
    """Converts a file:// URI to a proper file system path."""
    if "file" in uri:
        parsed = urllib.parse.urlparse(uri)

        # Handle Windows UNC paths (file://server/share/file.txt → \\server\share\file.txt)
        if parsed.netloc:
            file_path = f"{parsed.netloc}{parsed.path}"
        else:
            file_path = parsed.path

        # Strip leading slash for Windows drive letters (e.g., /C:/path → C:/path)
        if (
            os.name == "nt"
            and file_path.startswith("/")
            and len(file_path) > 2
            and file_path[2] == ":"
        ):
            file_path = file_path[1:]

        # Normalize for the OS (Windows backslashes, Unix forward slashes)
        return os.path.normpath(file_path)
    else:
        raise ValueError(f"{uri} is not a valid file uri")


def iter_files(directory: str, filter_str: str) -> Generator:
    """
    Iterate over files in a directory based on a specified filter pattern.

    Args:
        directory (str): The path to the directory.
        filter_str (str): The filter pattern to apply when filtering files.

    Returns:
        Generator: A generator yielding file paths that match the filter criteria.

    """
    folder_path = Path(directory)
    if folder_path.is_dir():
        # Using glob to filter files based on a pattern
        yield folder_path.glob(filter_str)



#  Copyright (c) 2023. OCX Consortium https://3docx.org. See the LICENSE

import pytest

from ocx_common.clients.clients import RestClient
from tests.conftest import MOCK_URL
from tests.mock_rest_responses import mock_server

client = RestClient()


class TestRestClient:
    @pytest.mark.commit(reason="Implement mock server")

    def test_rest_get_200(self, mock_server):
        url = f'{MOCK_URL}/get_200'
        response = client.get(url=url)
        user_data = response.json()
        assert response.status_code == 200
        assert user_data['id'] == 1
        assert user_data['name'] == 'John Doe'

    @pytest.mark.commit(reason="Implement mock server")
    def test_rest_get_400(self, mock_server):
        url = f'{MOCK_URL}/get_400'
        response = client.get(url=url)
        assert response.status_code == 400
        assert response.json()['error'] == 'Bad Request'

    @pytest.mark.commit(reason="Implement mock server")
    def test_post_201(self, mock_server):
        url = f'{MOCK_URL}/post_201'
        payload = {'id': 1,
                   'name': 'John Doe'}
        response = client.post(url=url, payload=payload)
        user_data = response.json()
        # Assertions to verify the response handling
        assert response.status_code == 201
        assert user_data['id'] == 1
        assert user_data['name'] == 'John Doe'



#  Copyright (c) 2025. #  OCX Consortium https://3docx.org. See the LICENSE
from pathlib import Path

import pytest

from ocx_common.utilities.validation import URIValidator
from tests.conftest import MODEL_FOLDER, TEST_MODEL


def test_validate_and_existence(shared_datadir):
    uri = shared_datadir / MODEL_FOLDER / TEST_MODEL
    validator = URIValidator(str(uri.resolve()))
    assert validator.is_valid(check_source=True)

@pytest.mark.parametrize(
    "uri, expected",
    [
        ("C:/Users/oca/AppData/Local/Temp/pytest-of-OCA/pytest-430/test_validate_10/data/models/NAPA-OCX_M1.3docx", True),  # ✅ Valid Windows absolute path
        ("https://3docx.org/fileadmin//ocx_schema//V300//OCX_Schema.xsd", True),  # ✅ A valid uri
        ("file://.path.txt", False),  # ❌  Windows local path
        ("https:3docx.org/fileadmin//ocx_schema//V300//OCX_Schema.xsd", False),  # ❌ Missing slashes
        ("file://C|/invalid/path.txt", False),  # ❌ Incorrect Windows drive format
        ("file:\\backslashes\\wrong.txt", False),  # ❌ Backslashes are not allowed
    ],
)
def test_validate(uri, expected):
    validator = URIValidator(uri)
    assert validator.is_valid() == expected

@pytest.mark.parametrize(
    "uri, expected",
    [
        ("C:/Users/oca/AppData/Local/Temp/pytest-of-OCA/pytest-430/test_validate_10/data/models/NAPA-OCX_M1.3docx", True),  # ✅ Valid Windows absolute path
        ("file://.path.txt", False),  # ❌  Local, but not absolute path
        ("https:3docx.org/fileadmin//ocx_schema//V300//OCX_Schema.xsd", False),  # ❌ Remote
        ("file://C:/users/path.txt", True),  # ✅  Local
    ],
)
def test_is_local_file(uri, expected):
    validator = URIValidator(uri)
    assert validator.is_local_file() == expected



#  Copyright (c) 2024. #  OCX Consortium https://3docx.org. See the LICENSE
"""A python XPath implementation for OCX types"""

# System imports
from typing import Any, Callable, List

# Third party imports
from lxml import etree
from lxml.etree import QName, XPathError

# Project imports
from ocx_common.x_path.xelement import LxmlElement


class OcxPathError(ValueError, XPathError):
    pass


class OcxPathBuilder:
    @staticmethod
    def select_all_named_nodes(nodename: str, namespace: str = "ocx") -> str:
        return f"//{namespace}:{nodename}"

    @staticmethod
    def select_current_node() -> str:
        return "."

    @staticmethod
    def select_parent_node() -> str:
        return "parent::*"

    @staticmethod
    def select_named_parent_node(parent: str, namespace: str = "ocx") -> str:
        return f"parent::{namespace}:{parent}"

    @staticmethod
    def select_named_nodes(node_name: str, namespace: str = "ocx") -> str:
        return f"//{namespace}:{node_name}"

    @staticmethod
    def select_any_nodes_with_global_attribute_name(
        attribute_name: str, namespace: str = "ocx"
    ) -> str:
        return f"//{namespace}:*[@{namespace}:{attribute_name}]"

    @staticmethod
    def select_any_nodes_with_attribute_value(
        attribute_name: str, attribute_value: str, namespace: str = "ocx"
    ) -> str:
        return f'//{namespace}:*[@{namespace}:{attribute_name}="{attribute_value}"]'

    @staticmethod
    def select_named_nodes_with_global_attribute_name(
        node_name: str, attribute_name: str, namespace: str = "ocx"
    ) -> str:
        return f"//{namespace}:{node_name}[@{namespace}:{attribute_name}]"

    @staticmethod
    def select_named_nodes_with_global_attribute_value(
        node_name: str,
        attribute_name: str,
        attribute_value: str,
        namespace: str = "ocx",
    ) -> str:
        return f'//{namespace}:{node_name}[@{namespace}:{attribute_name}="{attribute_value}"]'

    @staticmethod
    def select_any_nodes_with_local_attribute_name(
        attribute_name: str, namespace: str = "ocx"
    ) -> str:
        return f"//{namespace}:*[@{attribute_name}]"

    @staticmethod
    def select_any_nodes_with_local_value(
        attribute_name: str, attribute_value: str, namespace: str = "ocx"
    ) -> str:
        return f'//{namespace}:*[@{attribute_name}="{attribute_value}"]'

    @staticmethod
    def select_named_nodes_with_local_attribute_name(
        node_name: str, attribute_name: str, namespace: str = "ocx"
    ) -> str:
        return f"//{namespace}:{node_name}[@{attribute_name}]"

    @staticmethod
    def select_named_nodes_with_local_attribute_value(
        node_name: str,
        attribute_name: str,
        attribute_value: str,
        namespace: str = "ocx",
    ) -> str:
        return f'//{namespace}:{node_name}[@{attribute_name}="{attribute_value}"]'


class OcxPath:
    def __init__(
        self,
        document_root: etree.Element,
        namespaces: Any,
        extensions: Any = None,
        regexp: bool = True,
        smart_strings: bool = True,
    ):
        self._root = document_root
        self._namespaces = namespaces
        self._extensions = extensions
        self._regexp = regexp
        self._smart_strings = smart_strings

    def get_ocx_attribute_value_collection(
        self,
        element: etree.Element,
        attribute_name: str,
        namespace: str = "ocx",
    ) -> List[Any]:
        search = etree.XPath(
            path=OcxPathBuilder.select_any_nodes_with_global_attribute_name(
                attribute_name=attribute_name, namespace=namespace
            ),
            namespaces=self._namespaces,
            regexp=self._regexp,
            smart_strings=self._smart_strings,
        )
        result = search(element)
        return [v.get(attribute_name) for v in result]

    def get_ocx_attribute_with_value(
        self,
        element: etree.Element,
        attribute_name: str,
        attribute_value: str,
        namespace: str = "ocx",
    ) -> List[Any]:
        search = etree.XPath(
            path=OcxPathBuilder.select_any_nodes_with_attribute_value(
                attribute_name=attribute_name,
                attribute_value=attribute_value,
                namespace=namespace,
            ),
            namespaces=self._namespaces,
            regexp=self._regexp,
            smart_strings=self._smart_strings,
        )
        return search(element)

    def get_all_named_ocx_elements(
        self, name: str, namespace: str = "ocx"
    ) -> List[Any]:
        search = etree.XPath(
            path=OcxPathBuilder.select_all_named_nodes(
                nodename=name, namespace=namespace
            ),
            namespaces=self._namespaces,
            regexp=self._regexp,
            smart_strings=self._smart_strings,
        )
        return search(self._root)

    def get_all_named_children(
        self, node: etree.Element, child_name: str, namespace: str = "ocx"
    ) -> List[Any]:
        search = etree.XPath(
            path=OcxPathBuilder.select_all_named_nodes(
                nodename=child_name, namespace=namespace
            ),
            namespaces=self._namespaces,
            regexp=self._regexp,
            smart_strings=self._smart_strings,
        )
        return search(node)


class OcxGuidRef:
    def __init__(
        self,
        element_node: etree.Element,
        namespaces: Any,
        extensions: Any = None,
        regexp: bool = True,
        smart_strings: bool = True,
    ):
        self._node = element_node
        self._namespaces = namespaces
        self._extensions = extensions
        self._regexp = regexp
        self._smart_strings = smart_strings

    def _get_guids(self, nodes: List) -> Any:
        guids = []
        for element in nodes:
            ns = QName(element).namespace
            if element.get(f"{LxmlElement.namespaces_decorate(ns)}refType") is None:
                guids.append(
                    element.get(f"{LxmlElement.namespaces_decorate(ns)}GUIDRef")
                )
        return guids

    def get_all_guids(self) -> Callable:
        search = etree.XPath(
            path=OcxPathBuilder.select_any_nodes_with_global_attribute_name(
                attribute_name="GUIDRef"
            ),
            namespaces=self._namespaces,
            extensions=self._extensions,
            regexp=self._regexp,
            smart_strings=self._smart_strings,
        )
        nodes = search(self._node)
        return self._get_guids(nodes)

    def get_child_guids(self, node_name: str, namespace: str = "ocx") -> Callable:
        search = etree.XPath(
            path=OcxPathBuilder.select_any_nodes_with_global_attribute_name(
                attribute_name="GUIDRef"
            ),
            namespaces=self._namespaces,
            extensions=self._extensions,
            regexp=self._regexp,
            smart_strings=self._smart_strings,
        )
        nodes = search(self._node)
        return self._get_guids(nodes)



#  Copyright (c) 2023. OCX Consortium https://3docx.org. See the LICENSE

from ocx_common.loader.loader import DeclarationOfOcxImport, DynamicLoader
from ocx_common.parser.parsers import MetaData
from tests.conftest import SCHEMA_VERSION


def test_import_ocx_286():
    declaration = DeclarationOfOcxImport("ocx", "2.8.6")
    assert DynamicLoader.import_module(declaration)

def test_import_ocx_300():
    declaration = DeclarationOfOcxImport("ocx", "3.0.0")
    assert DynamicLoader.import_module(declaration)

def test_import_class():
    declaration = DeclarationOfOcxImport("ocx", SCHEMA_VERSION)
    assert DynamicLoader.import_class(declaration, "Vessel")


def test_ref_type_name():
    class_name = "RefTypeValue"
    value = "OCX_VESSEL"
    declaration = DeclarationOfOcxImport("ocx", SCHEMA_VERSION)
    data_class = DynamicLoader.import_class(declaration, class_name)
    instance = getattr(data_class, value)
    assert instance.name == "OCX_VESSEL"


def test_data_class_instance_namespace():
    class_name = "Vessel"
    declaration = DeclarationOfOcxImport("ocx", SCHEMA_VERSION)
    data_class = DynamicLoader.import_class(declaration, class_name)()
    namespace = MetaData.namespace(data_class)
    repo_folder = SCHEMA_VERSION.replace(".", "")
    repo_folder = f"V{repo_folder}"
    assert (
        namespace
        == f"https://3docx.org/fileadmin//ocx_schema//{repo_folder}//OCX_Schema.xsd"
    )


def test_get_all_class_names():
    classes = DynamicLoader.get_all_class_names("ocx", "3.0.0")
    assert len(classes) == 504


def test_get_declaration():
    declaration = DeclarationOfOcxImport("ocx", SCHEMA_VERSION).get_declaration()
    assert declaration == "ocx.ocx_300.ocx_300"


def test_get_name():
    name = DeclarationOfOcxImport("ocx", SCHEMA_VERSION).get_name()
    assert name == "ocx"


def test_get_version():
    version = DeclarationOfOcxImport("ocx", SCHEMA_VERSION).get_version()
    assert version == SCHEMA_VERSION



#  Copyright (c) 2022-2023. OCX Consortium https://3docx.org. See the LICENSE

# Sys imports
import re
from typing import Any, Dict, List, Optional, Union

# Third party imports
from loguru import logger
from lxml import etree
from lxml.etree import Element, ElementTextIterator, QName


class LxmlElement:
    """A wrapper class for the lxml etree.Element class main functions."""

    @staticmethod
    def items(element: Element) -> List:
        """Gets element attributes, as a sequence. The attributes are returned in an arbitrary order.

        Args:
            element: The current etree.Element node

        Returns:
            A List of all attributes of ``element``

        """
        return element.items()

    @staticmethod
    def get_root(element: Element) -> Element:
        """Return Element root node of the document that contains this element.

        Args:
            element: The current etree.Element node

        Returns:
            The root Element node of the document for this element

        """
        et = element.getroottree()
        return et.getroot()

    @staticmethod
    def get_base(element: Element) -> str:
        """The base URI of the document holding the Element (the location of the document)

        Args:
            element: The current etree.Element node

        Returns:
            The base URI of the document or empty string if unknown

        """
        base = element.base
        if base is None:
            return ""
        else:
            return base

    @staticmethod
    def get_source_line(element: Element) -> int:
        """Original line number as found by the parser or None if unknown.

        Args:
            element: The current etree.Element node

        Returns:
            The line number of the ``element`` in the source

        """
        return element.sourceline

    @staticmethod
    def get_parent(element: Element) -> Element:
        """Returns the parent of this element or None for the root element.

        Args:
            element: The current ``etree.Element`` node

        Returns:
            The parent Element node of the document for this element or none if ``element`` is the root

        """
        return element.getparent()

    @staticmethod
    def get_children(element: Element) -> List[Element]:
        """Returns all direct children of the xml element

        Args:
            element: The XML parent node

        Returns:
            A List of all children of 'element' excluding the parent itself.

        """
        return element.findall(".//")

    @staticmethod
    def get_xml_attrib(element: Element) -> Dict:
        """The XML attributes of an element

        Args:
            element: The XML parent node

        Returns:
            A dictionary of ``(key,value)`` pairs of the element attributes

        """
        return element.attrib

    @staticmethod
    def get_localname(element: Element) -> str:
        """The local name (type) of an XML element

        Args:
            element: The XML parent node

        Returns:
            The element local name

        """
        qn = QName(element)
        return qn.localname

    @classmethod
    def unique_tag(cls, element: Element) -> Any:
        """The unique tag of an XML element: ``{namespace}name``

        Args:
            element: The XML parent node

        Returns:
            The element unique tag

        """
        name = element.get("name")
        ns = cls.get_namespace(element)
        return f"{LxmlElement.namespaces_decorate(ns)}{name}"

    @staticmethod
    def get_name(element: Element) -> Any:
        """The name of an XML element defined by the attribute ``name``

        Args:
            element: The XML parent node

        Returns:
            The element local name or None if no name

        """
        name = element.get("name")
        if name is None:
            name = LxmlElement.strip_namespace_prefix(element.get("ref"))
        return name

    @staticmethod
    def get_use(element: Element) -> Any:
        """Whether XML element is required or optional given by the attribute value ``use``

        Args:
            element: The XML parent node

        Returns:

            The element use as a string, either ``required`` or ``optional``

        """
        use = element.get("use")
        if use is None:
            return "opt."
        if use == "required":
            use = "req."
        return use

    @staticmethod
    def get_reference(element: Element) -> Any:
        """The referenced XML element defined by the attribute ``ref``

        Args:
            element: The XML parent node

        Returns:

            The referenced element including namespace prefix if any. None if there is no reference

        """
        return element.get("ref")

    @staticmethod
    def is_enumeration(element: Element) -> bool:
        """Whether the attribute is an enumeration or not

        Args:
            element: The XML parent node

        Returns:

            true if the attribute is an enumeratos, false otherwise

        """
        return LxmlElement.has_child_with_name(element, "enumeration")

    @staticmethod
    def is_reference(element: Element) -> bool:
        """Whether the element is a reference or not

        Args:
            element: The XML parent node

        Returns:

            true if the element is a reference, false otherwise

        """
        return element.get("use") is not None

    @staticmethod
    def is_mandatory(element: Element) -> bool:
        """The element use. True if required, False otherwise

        Args:
            element: The element node

        Returns:

            True if the element is mandatory, false otherwise

        """
        lower, upper = LxmlElement.cardinality(element)
        if lower == 0:
            return False
        else:
            return True

    @classmethod
    def cardinality_string(cls, element) -> str:
        """Return the element cardinality formatted string."""
        lower, upper = cls.cardinality(element)
        if upper == "unbounded":
            upper = "\u221e"  # UTF-8 Infinity symbol
        return f"[{lower}, {upper}]"

    @staticmethod
    def cardinality(element) -> tuple:
        """Establish the cardinality of the ``Element``

        Args:
            element: the ``etree.Element`` instance

        Returns:

            The tuple of the element lower and upper bounds (lower, upper)

        """

        attributes = LxmlElement.get_xml_attrib(element)
        if "minOccurs" in attributes:
            lower = int(attributes["minOccurs"])
        else:
            use = attributes.get("use")
            if use is not None:
                if use == "req" or use == "required":
                    lower = 1
                else:
                    lower = 0
            else:
                lower = 1
        if "maxOccurs" in attributes:
            upper = attributes["maxOccurs"]
        else:
            upper = 1
        # Find the closest xs:sequence or xs:choice ancestor which overrules mandatory use
        for item in element.iterancestors("{*}sequence", "{*}choice"):
            attributes = item.attrib
            if "minOccurs" in attributes:
                lower = int(attributes["minOccurs"])
            if "maxOccurs" in attributes:
                upper = attributes["maxOccurs"]
            break
        return lower, upper

    @staticmethod
    def is_choice(element) -> bool:
        """Return True if the element ancestor is an xs:choice, False otherwise

        Args:
            element: the etree.Element instance

        Returns:

            True if the element node is a choice, false otherwise

        """
        choice = False
        # Find the closest sequence or choice ancestor which overrules mandatory use
        for item in element.iterancestors("{*}sequence", "{*}choice"):
            qn = QName(item)
            if qn.localname == "choice":
                choice = True
            break
        return choice

    @staticmethod
    def is_substitution_group(element) -> bool:
        """Return True if the element is a substitution

        Args:
            element: the etree.Element instance

        Returns:

            True if the element is a  substitutionGroup, false otherwise

        """
        attributes = LxmlElement.get_xml_attrib(element)
        if "substitutionGroup" in attributes:
            return True
        else:
            return False

    @staticmethod
    def is_abstract(element) -> bool:
        """Return True if the element is abstract

        Args:
            element: the ``etree.Element`` instance

        Returns:

            True if the element abstract, false otherwise

        """
        attributes = LxmlElement.get_xml_attrib(element)
        if "abstract" in attributes:
            return True
        else:
            return False

    @staticmethod
    def get_substitution_group(element: Element) -> Optional[Any]:
        """Return the name of the element's substitutionGroup

        Args:
            element: the ``etree.Element`` instance

        Returns:
            name of substitutionGroup, None if no substitutionGroup

        """
        attributes = LxmlElement.get_xml_attrib(element)
        return attributes.get("substitutionGroup")

    @staticmethod
    def get_restriction(element: Element) -> str:
        """Return the element restriction

        Args:
            element: the ``etree.Element`` instance

        Returns:
            restriction type

        """
        restriction = ""
        for item in LxmlElement.iter(element, "{*}restriction"):
            restriction = item.get("base")
        return restriction

    @staticmethod
    def get_element_text(element: Element) -> str:
        """The text between the element's start and end tags without any tail text.

        Args:
            element: the etree.Element instance

        Returns:

            The element text stripped of any special characters

        """
        text = ""
        description = ""
        for text in ElementTextIterator(element, with_tail=False):
            description = description + text
            break
        text = re.sub("[\n\t\r]", "", description)  # Strip off special characters
        return text

    @staticmethod
    def get_namespace(element: Element) -> str:
        """The namespace of an XML element

        Args:
            element: The XML parent node

        Returns:

            The element namespace

        """
        qn = QName(element)
        return qn.namespace

    @staticmethod
    def iter(element: Element, tag=None, *tags) -> etree.ElementDepthFirstIterator:
        """Iterate over all elements in the subtree in document order (depth first pre-order),
            starting with this element.

            Can be restricted to find only elements with specific tags: pass ``{ns}localname`` as tag.
            Either or both of ns and localname can be * for a wildcard; ns can be empty for no namespace.
            ``localname`` is equivalent to ``{}localname`` (i.e. no namespace)
            but ``*`` is ``{*}*`` (any or no namespace), not ``{}*``.

            Passing multiple tags (or a sequence of tags) instead of a single tag
            will let the iterator return all elements matching any of these tags, in document order.

        Args:
            element: The etree.Element node to search from
            tag: The name of the child
        Returns:
            An iterator filtered by tags if specified.

        Example:
            .. highlight:: python
            .. code-block:: python

                for type in LxmlElement.iter(root, {*}complexType)
                    print(type.tag)

            will iterate over all ``complexType`` tags and print the tag starting from the document root .

        """
        return element.iter(tag, *tags)

    @staticmethod
    def find_all_children_with_name(
        element: Element, child_name: str, namespace: str = "*"
    ) -> List:
        """Find all the XML element's children with  name ``child_name``

        Args:
            element: The XML parent node to search from
            child_name: The name of the child
            namespace: The search namespace. Default is the wildcard ``*`` matching any namespace

        Returns:

            A list of elements. Empty list if no children can be found

        """
        xpath = f".//{LxmlElement.namespaces_decorate(namespace)}{child_name}"
        return element.findall(xpath)

    @staticmethod
    def find_child_with_name(
        element: Element, child_name: str, namespace: str = "*"
    ) -> Element:
        """Find the first direct child of the XML element's children with  name ``child_name``

        Args:
            element: The XML parent node to search from
            child_name: The name of the child
            namespace: The search namespace. Default is the wildcard '*' matching any namespace

        Returns:
            The child element as etree.Element. None if no child can be found

        """
        xpath = f".//{LxmlElement.namespaces_decorate(namespace)}{child_name}"
        return element.find(xpath)

    @staticmethod
    def find_attributes(element: Element, namespace: str = "*") -> List[Element]:
        """Find all sub elements of type xs:attribute

        Args:
            element: The XML parent node to search from
            namespace: The search namespace. Default is the wildcard ``*`` matching any namespace

        Returns:

            The list of the xs:attribute type found

        """
        xpath = f".//{LxmlElement.namespaces_decorate(namespace)}attribute"
        return element.findall(xpath)

    @staticmethod
    def find_attribute_groups(element: Element, namespace: str = "*") -> List[Element]:
        """Find all sub elements of type ``xs:attributeGroup``

        Args:
            element: The XML parent node to search from
            namespace: The search namespace. Default is the wildcard ``*`` matching any namespace

        Returns:
            Attribute groups

        """
        xpath = f".//{LxmlElement.namespaces_decorate(namespace)}attributeGroup"
        return element.findall(xpath)

    @staticmethod
    def has_child_with_name(
        element: Element, child_name: str, namespace: str = "*"
    ) -> bool:
        """Check if the element has a child with  name 'child_name'

        Args:
            element: The XML parent node to search from
            child_name: The name of the child
            namespace: The search namespace. Default is the wildcard ``*`` matching any namespace

        Returns:
            True if the element has a child with name ``child_name`` False otherwise

        """
        xpath = f".//{LxmlElement.namespaces_decorate(namespace)}{child_name}"
        return len(element.findall(xpath)) > 0

    @staticmethod
    def find_all_children_with_attribute_value(
        element: Element,
        name: str,
        attrib_name: str,
        attrib_value: str,
        namespace: str = "*",
    ) -> List:
        """Find all the XML elements with the attribute name 'attrib_name' having a given value  'attrib_value'

        Args:
            element: The XML parent node to search from
            name: The name of the element with attrib_name and attrib_value
            attrib_name: The name of the attribute
            attrib_value: The value of the attribute
            namespace: The search namespace. Default is the wildcard ``*`` matching any namespace

        Returns:
            All children having attributes with name ``attrib_name`` and value ``attrib_value``.
            Empty list if no children can be found

        """
        xpath = f'.//{LxmlElement.namespaces_decorate(namespace)}{name}[@{attrib_name}="{attrib_value}"]'
        return element.findall(xpath)

    @staticmethod
    def find_all_children_with_name_and_attribute(
        element: Element,
        child_name: str,
        attrib_name: str,
        namespace: str = "*",
    ) -> List:
        """Find all the XML elements with name ``name`` and attribute name ``attrib_name``

        Args:
            element: The XML parent node to search from
            child_name: The name of the children elements
            attrib_name: The name of the attribute
            namespace: The search namespace. Default is the wildcard matching any namespace


        Returns:
            All elements having attributes with name ``attrib_name``.
            An empty list if no children can be found

        """
        xpath = f".//{LxmlElement.namespaces_decorate(namespace)}{child_name}[@{attrib_name}]"
        return element.findall(xpath)

    @staticmethod
    # Finding the assertion in the code.
    def find_assertion(element: Element, namespace="*") -> Union[str, None]:
        """Find any assertions under the ``element``

        Args:
            element: The XML parent node to search from
            namespace: The search namespace. Default is the wildcard ``*`` matching any namespace

        Returns:
            The assertion test as a string.
            None if no assertion tag is found

        """
        test = None
        # Assertions
        xpath = f".//{LxmlElement.namespaces_decorate(namespace)}{'assert'}"
        asserts = element.findall(xpath)
        if len(asserts) > 0:
            attrib = asserts[0].attrib
            if "test" in attrib:
                test = attrib["test"]
        return test

    @staticmethod
    def namespace_prefix(element: str) -> Union[str, None]:
        """Returns the namespace prefix of an element if any

        Args:

            element: The element name with or without prefix as a string

        Returns:

            The element prefix string or None if no prefix

        """
        i = element.find(":")
        if not i == -1:
            element = element[0:i]
            return element
        return None

    @staticmethod
    def replace_ns_tag_with_ns_prefix(element: str, namespaces: Dict) -> str:
        """Replace the namespace tag with a mapped namespace prefix.

        Args:

            element: The element name with or without prefix as a string
            namespaces: the namespace tag to prefix mapping

        Returns:

            the element with prefix

        """

        nsprefix = list(iter(namespaces))
        nstags = list(iter(namespaces.values()))
        qn = QName(element)
        prefix = ""
        try:
            index = nstags.index(qn.namespace)
            prefix = nsprefix[index]
            if prefix is None:
                nstags.pop(index)
                nsprefix.pop(index)
                index = nstags.index(qn.namespace)
                prefix = nsprefix.index(index)
        except ValueError as e:
            logger.error(f"{qn.namespace} is not in the namespace list: {e}")
        if prefix == "":
            logger.debug(f"Empty namespace prefix in element {qn.localname}")
        return f"{prefix}:{qn.localname}"

    @staticmethod
    def namespaces_decorate(ns: str) -> str:
        """Decorate a string with curly brackets to form a valid XML namespace

        Args:
            ns: The namespace

        Returns:
            The curly decorated namespace

        """
        return "{" + ns + "}"

    @staticmethod
    def strip_namespace_prefix(element: str) -> str:
        """Returns the element name without the namespace prefix

        Args:
            element: The element name with or without prefix as a string

        Returns:
            The element without namespace prefix

        """
        i = element.find(":")
        if not i == -1:
            element = element[i + 1 : len(element)]
        return element

    @staticmethod
    def strip_namespace_tag(element: str) -> str:
        """Returns the element name without the namespace tag

        Args:
            element: The element name with or without namespace as a string

        Returns:
            The element without namespace tag

        """
        i = element.find("}")
        if not i == -1:
            element = element[i + 1 : len(element)]
        return element



#  Copyright (c) 2025. #  OCX Consortium https://3docx.org. See the LICENSE
import re
from pathlib import Path
from typing import Dict
from loguru import logger

from ocx_common.utilities.validation import URIValidator


class OcxXml:
    """Find the schema version of an 3Docx XML model."""

    @staticmethod
    def get_version(model: str) -> str:
        """
        The schema version of the model.
        Args:
            model: The source file path or uri

        Returns:
            The schema version of the 3Docx XML model or None if the file cannot be found.
        """
        version = None
        validator = URIValidator(model)
        if validator.is_local_file():
            content = Path(model).read_text().split()
            for item in content:
                if "schemaVersion" in item:
                    version = item[item.find("=") + 2 : -1]
        else:
            logger.error(f"The uri {model} is not a file path")
        return version

    @staticmethod
    def get_ocx_namespace(model: str) -> str:
        """Return the OCX schema namespace of the model.

        Args:
            model: The source path or uri

        Returns:
              The OCX schema namespace of the model.
        """
        namespace = "NA"
        ocx_model = Path(model).resolve()
        if OcxXml.has_ocx_namespace(str(ocx_model)):
            content = ocx_model.read_text().split()
            for item in content:
                if "xmlns:ocx" in item:
                    # Extract all characters between double quotes
                    namespace = re.findall(r'"(.*?)"', item)
                    namespace = namespace[0]
        return namespace

    @staticmethod
    def has_ocx_namespace(model: str) -> bool:
        """Return True if the OCX schema namespace is defined.

        Args:
            model: The source path or uri

        Returns:
              True if the xmlns:ocx is defined, False otherwise.
        """
        ocx_model = Path(model).resolve()
        content = ocx_model.read_text()
        return "xmlns:ocx" in content

    @staticmethod
    def has_unitsml_namespace(model: str) -> bool:
        """Return True if the OCX schema unitsml namespace is defined.

        Args:
            model: The source path or uri

        Returns:
              True if the xmlns:unitsml is defined, False otherwise.
        """
        ocx_model = Path(model).resolve()
        content = ocx_model.read_text()
        return "xmlns:unitsml" in content

    @staticmethod
    def get_all_namespaces(model: str) -> Dict:
        """Return all the xmlns namespace map defined in the 3Docx model.

        Args:
            model: The source path or uri

        Returns:
              The namespace mappings.
        """
        namespaces = {}
        ocx_model = Path(model).resolve()
        content = ocx_model.read_text().split()
        for item in content:
            if "xmlns:" in item:
                if match := re.search(r':([^=]+)="([^"]+)"', item):
                    prefix = match[1]
                    namespace = match[2]
                    namespaces[prefix] = namespace
        return namespaces



#  Copyright (c) 2023. OCX Consortium https://3docx.org. See the LICENSE
"""Module for parsing a 3Docx model."""

# system imports
from abc import ABC
from enum import Enum
from pathlib import Path

from typing import TypeVar, Dict, Optional
from dataclasses import dataclass, is_dataclass

# 3rd party imports
import lxml.etree
from loguru import logger
from lxml.etree import Element
from xsdata.exceptions import ParserError
from xsdata.formats.dataclass.context import XmlContext, XmlContextError
from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.parsers.config import ParserConfig
from xsdata.formats.dataclass.parsers.handlers import LxmlEventHandler

# Project imports
from ocx_common.interfaces.interfaces import IObservable
from ocx_common.loader.loader import DeclarationOfOcxImport, DynamicLoader
from ocx_common.utilities.ocx_xml import OcxXml
from ocx_common.utilities.validation import URIError, URIValidator


T = TypeVar("T")


class OcxParserError(ParserError):
    """Errors raised by this module."""


class ObservableEvent(Enum):
    """Events that can be listened to and broadcast."""

    DATACLASS = "dataclass"
    REPORT = "report"
    SERIALIZE = "serialize"


class XmlParserError(ValueError):
    """Parser errors."""


class MetaData:
    """Dataclass metadata."""

    @staticmethod
    def meta_class_fields(data_class: dataclass) -> Dict:
        """
        Return the dataclass metadata.

        Args:
            data_class: The dataclass instance

        Returns:
            The metadata of the class
        """
        return dict(data_class.Meta.__dict__.items())

    @staticmethod
    def class_name(data_class: dataclass) -> str:
        """Return the name of the class"""
        declaration = str(data_class.__class__)
        return declaration[declaration.rfind(".") + 1 : -2]

    @staticmethod
    def namespace(data_class: dataclass) -> str:
        """Get the OCX namespace

        Args:
            data_class: The dataclass instance

        Returns:
            The namespace of the dataclass
        """
        return MetaData.meta_class_fields(data_class).get("namespace")

    @staticmethod
    def name(data_class: dataclass) -> str:
        """Get the OCX name.

        Args:
            data_class: The dataclass instance

        Returns:
            The name of the OCX type
        """
        return MetaData.meta_class_fields(data_class).get("name")


class OcxNotifyParser(IObservable, ABC):
    """Ocx notification parser class for 3Docx XML files.

    Args:
        fail_on_unknown_properties: Don't bail out on unknown properties.
        fail_on_unknown_attributes: Don't bail out on unknown attributes
        fail_on_converter_warnings: bool = Convert warnings to exceptions
    """

    def __init__(
        self,
        fail_on_unknown_properties: bool = False,
        fail_on_unknown_attributes: bool = False,
        fail_on_converter_warnings: bool = True,
    ):
        context = XmlContext()
        parser_config = ParserConfig(
            fail_on_unknown_properties=fail_on_unknown_properties,
            fail_on_unknown_attributes=fail_on_unknown_attributes,
            fail_on_converter_warnings=fail_on_converter_warnings,
            class_factory=self.class_factory,
        )
        self._parser = XmlParser(config=parser_config, context=context)
        self._subscribers = set()

    def subscribe(self, observer):
        self._subscribers.add(observer)
        return

    def unsubscribe(self, observer):
        self._subscribers.remove(observer)
        return

    def update(self, event: ObservableEvent, payload: Dict):
        for observer in self._subscribers:
            observer.update(event, payload)

    def class_factory(self, clazz, params):
        """Custom class factory method"""
        name = clazz.__name__
        new_data_class = clazz(**params)
        # Broadcast an update
        namespace = MetaData.namespace(clazz)
        # name = MetaData.name(clazz)
        fields = MetaData.meta_class_fields(clazz)
        logger.debug(f"Meta fields: {fields}")
        tag = "{" + namespace + "}" + name
        self.update(ObservableEvent.DATACLASS, {"name": tag, "object": new_data_class})
        return new_data_class

    def parse(self, xml_file: str) -> Optional[T]:
        """Parse a 3Docx XML model and return the root dataclass.

        Args:
            xml_file: The 3Docx xml file or url to parse.

        Returns:
            The root dataclass instance of the parsed 3Docx XML.
        """
        try:
            validator = URIValidator(xml_file)
            if validator.is_local_file():
                tree = lxml.etree.parse(xml_file)
                root = tree.getroot()
                version = OcxXml.get_version(xml_file)
                declaration = DeclarationOfOcxImport("ocx", version)
                # Load target schema version module
                ocx_module = DynamicLoader.import_module(declaration)
                ocx_xml = self._parser.parse(root, ocx_module.OcxXml)
                if not is_dataclass(ocx_xml):
                    logger.error(f"{ocx_xml.__class_name__} is not a data class")
                    return None
                return ocx_xml
            else:
                logger.error(f'The uri r"{xml_file}" is not a local file')
        except (
            lxml.etree.XMLSyntaxError,
            ImportError,
            XmlContextError,
            ParserError,
        ) as e:
            logger.error(e)
            raise XmlParserError(e) from e

    def parse_element(self, element: Element, ocx_module) -> Optional[T]:
        """Parse a 3Docx XML element and return the dataclass.

        Args:
            element: The 3Docx XML Element to parse.

        Returns:
            The element dataclass instance if created, else returns None.
        """
        try:
            cls = self._parser.parse(element, ocx_module)
            if not is_dataclass(cls):
                logger.error(f"{cls.__name__} is not a dataclass")
            return cls
        except (
            lxml.etree.XMLSyntaxError,
            ImportError,
            XmlContextError,
            ParserError,
        ) as e:
            logger.error(e)
            raise XmlParserError(e) from e


class OcxParser:
    """OcxParser class for 3Docx XML files.

    Args:
        ocx_model: The file path or URL to the source model.

    """

    def __init__(
        self,
        ocx_model: str,
    ):
        self._parser = XmlParser(handler=LxmlEventHandler)
        self._tree: lxml.etree = None
        self._version: str = ""
        try:
            validator = URIValidator(ocx_model)
            if validator.is_valid(check_source=True):
                file = Path(ocx_model).resolve()
                self._version = OcxXml.get_version(str(file))
                self._tree = lxml.etree.parse(file)
        except (URIError, ParserError) as e:
            logger.error(e)
            raise OcxParserError(e) from e

    def get_root(self) -> lxml.etree.Element:
        """Return the document root"""
        return self._tree.getroot()



#  Copyright (c) 2023. OCX Consortium https://3docx.org. See the LICENSE
"""Module for server clients."""

# System imports
import json
from abc import ABC
from enum import Enum
from io import BytesIO
from typing import Dict

import pycurl
import requests

# Third party
from loguru import logger
from pycurl import error as pycurl_error
from requests.models import Response

# Project imports
from ocx_common.interfaces.interfaces import IRestClient


class RequestClientError(requests.RequestException):
    """Request client errors."""


class CurlClientError(pycurl_error):
    """Curl client errors."""


class EmbeddingMethod(Enum):
    """Embedding type."""

    STRING = "STRING"
    URL = "URL"
    BASE64 = "BASE64"


class RestClient(IRestClient, ABC):
    """
    Request client
    """

    def __init__(self, headers: Dict = None, timeout: int = 30):
        if headers is None:
            headers = {
                "Accept": "application/json",
                "Content-Type": "application/json",
            }
        super().__init__(headers, timeout)
        self._session = requests.Session()

    def get(self, url: str, headers: dict = None) -> Response:
        """
            Get method.

        Args:
            headers: The request header dictionary
            url: The resource url

        Returns:
            The request Response object.

        Raises:
             RequestClientError on (HTTPError, ConnectionError).
        """

        try:
            if headers is None:
                headers = self._headers
            response = self._session.get(url, headers=headers, timeout=self._timeout)
            return response
        except (
            requests.exceptions.HTTPError,
            requests.exceptions.ConnectionError,
        ) as err:
            logger.error(err)
            raise RequestClientError(err) from err

    def post(self, url: str, payload: Dict, headers: Dict = None) -> Response:
        """
            Post method.
        Args:
            headers: The request headers
            url: the resource.
            payload: The request body.

        Returns:
            The Response object.

        Raises:
              RequestClientError on (HTTPError, ConnectionError).
        """

        try:
            if headers is None:
                headers = self._headers
            response = self._session.post(
                url, headers=headers, json=payload, timeout=self._timeout
            )
            return response
        except (
            requests.exceptions.HTTPError,
            requests.exceptions.ConnectionError,
        ) as err:
            logger.error(err)
            raise RequestClientError(err) from err

    def set_headers(self, headers: Dict):
        """
            Set the request headers
        Args:
            headers: The headers declaration

        """

        self._headers = headers

    def put(self, url: str, payload: Dict, headers: Dict = None) -> Response:
        """
            Put method.
        Args:
            headers: Request header
            url: the resource.
            payload: The request body.

        Returns:
            The Response object.

        Raises:
              RequestClientError on (HTTPError, ConnectionError).

        """

        raise RequestClientError("put method not implemented")

    def patch(self, url: str, payload: Dict, headers: Dict = None) -> Response:
        """
            Patch method.
        Args:
            headers: The request header
            url: the resource.
            payload: The request body.

        Returns:
            The Response object.

        Raises:
              RequestClientError on (HTTPError, ConnectionError).
        """

        raise RequestClientError("patch method not implemented")

    def delete(self, url: str, payload: Dict, headers: Dict = None) -> Response:
        """
            Delete method.
        Args:
            headers: The request header
            url: the resource.
            payload: The request body.

        Returns:
            The Response object.

        Raises:
              RequestClientError on (HTTPError, ConnectionError).
        """

        raise RequestClientError("delete method not implemented")


class CurlRestClient(IRestClient, ABC):
    """
    cURL client
    """

    def __init__(self, headers: Dict = None, timeout: int = 30):
        super().__init__(headers, timeout)

    def pycurl_to_requests_response(
        self, buffer: BytesIO, status_code: int
    ) -> Response:
        """A function to convert pycurl response to a requests Response object"""
        # Create a requests Response object
        response = Response()
        response._content = buffer.getvalue()  # Set the content of the response
        response.status_code = status_code  # Set the status code
        return response

    def get(self, url: str, headers: dict = None) -> Response:
        """
            Get method.

        Args:
            headers: Request header
            url: The resource endpoint

        Returns:
            The Response object.

        Raises:
             CurlClientError if the status code is not 200.
        """

        curl = pycurl.Curl()
        curl.setopt(pycurl.URL, url)
        curl.setopt(pycurl.TIMEOUT, self._timeout)
        buffer = BytesIO()
        # Set the headers
        curl_headers = []
        if headers is None:
            for k in self._headers:
                curl_headers.append(f"{k}:{self._headers[k]}")
        curl.setopt(pycurl.HTTPHEADER, headers)
        # Set the write function to store the response in the buffer
        curl.setopt(pycurl.WRITEDATA, buffer)
        try:
            # Perform the request
            curl.perform()
            # Get the status code
            status_code = curl.getinfo(pycurl.RESPONSE_CODE)
            # Get the response
            response = self.pycurl_to_requests_response(buffer, status_code)
            curl.close()
            return response
        except pycurl.error as e:
            msg = f"Failed to get data from {url}: Response: {e}"
            logger.error(msg)
            raise CurlClientError(msg) from e

    def post(self, url: str, payload: Dict, headers: Dict = None) -> Response:
        """
            Post method.
        Args:
            headers: The request headers
            url: the resource.
            payload: The request body.

        Returns:
            The Response object.

        Raises:
              RequestClientError on pycurl.error.
        """

        curl = pycurl.Curl()
        curl.setopt(pycurl.URL, url)
        # curl.setopt(pycurl.TIMEOUT, self._timeout)
        body = json.dumps(payload)
        buffer = BytesIO()
        # Set the HTTP method to POST
        curl.setopt(pycurl.POST, 1)
        # Set the POST data
        curl.setopt(pycurl.POSTFIELDS, body)
        # Set the headers
        if headers is None:
            headers = self._headers
        curl.setopt(pycurl.HTTPHEADER, headers)
        # Create a BytesIO object to store the header data
        header_buffer = BytesIO()
        # Set the WRITEFUNCTION to write the header data to the BytesIO object
        curl.setopt(pycurl.HEADERFUNCTION, header_buffer.write)
        # Set the write function to store the response in the buffer
        curl.setopt(pycurl.WRITEFUNCTION, buffer.write)
        try:
            # Perform the request
            curl.perform()
            status_code = curl.getinfo(pycurl.HTTP_CODE)
            response = self.pycurl_to_requests_response(buffer, status_code)
            curl.close()
            return response
        except pycurl.error as e:
            msg = f"Failed to post data to {url}: Response: {e}"
            logger.error(msg)
            raise CurlClientError(msg) from e

    def set_headers(self, headers: Dict):
        """
        Set the request headers
        Args:
            headers: The headers declaration
        """
        self._headers = [f"{k}:{v}" for k, v in headers.items()]

    def put(self, url: str, payload: Dict, headers: Dict = None) -> Response:
        """
            Put method.
        Args:
            headers: Request header
            url: the resource.
            payload: The request body.

        Returns:
            The Response object.

        Raises:
              RequestClientError on (HTTPError, ConnectionError).

        """

        raise CurlClientError("put method not implemented")

    def patch(self, url: str, payload: Dict, headers: Dict = None) -> Response:
        """
            Patch method.
        Args:
            headers: The request header
            url: the resource.
            payload: The request body.

        Returns:
            The Response object.

        Raises:
              RequestClientError on (HTTPError, ConnectionError).
        """

        raise CurlClientError("patch method not implemented")

    def delete(self, url: str, payload: Dict, headers: Dict = None) -> Response:
        """
            Delete method.
        Args:
            headers: The request header
            url: the resource.
            payload: The request body.

        Returns:
            The Response object.

        Raises:
              RequestClientError on (HTTPError, ConnectionError).
        """

        raise CurlClientError("delete method not implemented")



#  Copyright (c) 2023-2025. OCX Consortium https://3docx.org. See the LICENSE
import re
from pathlib import Path
from typing import Optional

from loguru import logger
from xsdata.codegen import opener

# Third part imports
from xsdata.utils.downloader import Downloader

from ocx_common.decorators.decorators import exception_handler

# Module imports
from ocx_common.utilities.validation import URIValidator


def set_schema_location_to_local_file(xsd_content: str) -> str:
    pattern = r'schemaLocation="([^"]+)"'  # Capture the value inside schemaLocation=""
    match = re.search(pattern, xsd_content)  # Find the first occurrence
    if match:
        schema_location = match.group(1)  # Return the extracted schemaLocation value
        schema_file = Path(schema_location).name
        # Replace it with the new local location
        new_location = f"file://.{schema_file}"
        updated_xsd = re.sub(
            pattern, f'schemaLocation="{new_location}"', xsd_content, count=1
        )
        return updated_xsd
    else:
        logger.debug("No schema location found")
        return xsd_content


class SchemaDownloader(Downloader):
    """Downloader specialisation class.

    Arguments:
        output: The location of the download folder relative to current directory

    Args:
        folder: The download target folder.

    Properties:
        schema_folder: The path to the schema download folder
        change_schema_location: Replace the ``schemaLocation`` attribute with the new schema location.
    """

    def __init__(self, folder: Path, change_schema_location: bool = True):
        super().__init__(folder)
        self.schema_folder = folder
        self.change_schema_location = change_schema_location

    def write_file(self, uri: str, location: Optional[str], content: str):
        """
        Override super class method and output all schemas into one folder.

        Arguments:
            content: The schema content.
            location: The download location.
            uri: the download target resource. All referenced schemas will be collected.
        """
        # Get the uri file name
        name = Path(uri).name
        file_path = self.schema_folder / name
        file_path.write_text(content, encoding="utf-8")
        logger.debug(
            f"Writing schema {file_path.resolve()} to folder {self.schema_folder.resolve()}"
        )
        # logger.debug(content)
        self.downloaded[uri] = file_path

        if location:
            self.downloaded[location] = file_path

    @exception_handler(BaseException)
    def wget(self, uri: str, location: Optional[str] = None):
        """Download handler for any uri input with circular protection.
        Override super class method to handle a local file.

        """
        validator = URIValidator(uri)
        if validator.is_valid():
            if not (
                uri in self.downloaded or (location and location in self.downloaded)
            ):
                self.downloaded[uri] = None
                self.downloaded[location] = None
                self.adjust_base_path(uri)

                logger.info(f"Fetching {uri}")
                if validator.is_local_file():
                    with open(str(uri), "rb") as file:
                        input_stream = file.read()
                else:
                    input_stream = opener.open(uri).read()  # nosec
        else:
            input_file = Path(uri).resolve()
            with open(str(input_file), "rb") as file:
                input_stream = file.read()
        if uri.endswith("wsdl"):
            self.parse_definitions(uri, input_stream)
        else:
            self.parse_schema(uri, input_stream)
            # replace the schema location
            content = input_stream.decode()
            if self.change_schema_location:
                content = set_schema_location_to_local_file(content)
            self.write_file(uri, location, content)



#  Copyright (c) 2024. OCX Consortium https://3docx.org. See the LICENSE

import os
from pathlib import Path

import pytest

# Project import
from ocx_common.utilities import utilities

from .conftest import MODEL_FOLDER, TEST_MODEL


def test_is_substring_in_list_1():
    assert utilities.is_substring_in_list(
        "Doe", ["John Doe", "Alicia", "Hello world", "package.file"]
    )


def test_is_substring_in_list_2():
    assert utilities.is_substring_in_list(
        ".file", ["John Doe", "Alicia", "Hello world", "package.file"]
    )


def test_is_substring_in_list_3():
    assert not utilities.is_substring_in_list(
        "Help", ["John Doe", "Alicia", "Hello world", "package.file"]
    )


def test_parent_directory(shared_datadir):
    file = shared_datadir / TEST_MODEL
    assert utilities.parent_directory(str(file.resolve())) == str(shared_datadir.resolve())


def test_all_equal_1():
    assert utilities.all_equal(["John", "John", "John", "John", "John", "John", "John"])


def test_all_equal_2():
    assert not utilities.all_equal(["John", "Alice", "John", "John", "John", "John", "John"])


def test_camel_case_split():
    assert utilities.camel_case_split("CamelCase") == ["Camel", "Case"]


def test_dromedary_case_split():
    assert utilities.dromedary_case_split("dromedaryCase") == ["dromedary", "Case"]


def test_list_files_in_directory(shared_datadir):
    folder = shared_datadir / MODEL_FOLDER
    assert utilities.list_files_in_directory(folder, "*.3docx") == [TEST_MODEL]


def test_resource_path(shared_datadir):
    file = Path.joinpath(shared_datadir, TEST_MODEL)
    path = utilities.resource_path(str(file))
    assert path == str(file)


def test_get_file_path(shared_datadir):
    file = Path.joinpath(shared_datadir, TEST_MODEL)
    path = utilities.get_file_path(str(file))
    assert path == str(file)

@pytest.mark.parametrize(
    "path, expected",
    [
        ("C:/Users/oca/AppData/Local/Temp/pytest-of-OCA/pytest-430/test_validate_10/data/models/NAPA-OCX_M1.3docx", True),  # ✅ Valid Windows absolute path
        ("C://Users/oca//NAPA-OCX_M1.3docx", True),  # ✅ Valid Windows absolute path with double slashes
        ("D://Users/oca//NAPA-OCX_M1.3docx", True),  # ✅ Valid Windows absolute path with double slashes
        ("..\relative\path\file.txt", False), # ❌ ( relative path)
        ("C:Windows/System32/", False), # ❌  (Missing / after drive)
        ("K/Windows/System32/", False),  # ❌  (Missing : after drive)
    ],
)

def test_is_valid_absolute_windows_path(path, expected):
    assert utilities.is_valid_absolute_windows_path(path) == expected


@pytest.mark.parametrize(
    "path, expected",
    [
        ("/home/user/file.txt", True),  # ✅ Valid absolute path
        ("/var/log/system.log", True),  # ✅ Valid absolute
        ("./relative/path/to/file", True),  # ✅ Valid relative path
        ("../parent/directory/script.sh", True),  # ✅ Valid relative path
        ("~/Documents/file.pdf", True), # ✅ Valid home directory path
        ("/usr/local/bin/", True), # ✅Valid directory path
        (".hiddenfile", True), # ✅Valid hidden file
        ("/tmp/my_file-123.log", True), # ✅Valid tmp file
        ("home/user/file.txt", False),  # ❌ Missing leading `/`, `./`, or `../`
        ("/invalid|name/file.txt", False),  # ❌ Contains `|`
        ("/etc/passwd?query=1", False),  # ❌ Contains `?`
        ("C:\\Windows\\System32\\cmd.exe", False),  # ❌ Windows path
    ],
)

def test_valid_unix_paths(path, expected):
    assert(utilities.is_valid_unix_file_path(path)) == expected

@pytest.mark.parametrize(
    "uri, expected",
    [
        ("file://localhost:80/home/user/file.txt", False),  # ❌ remote file path
        ("file://C:/user/system.log", True),  # ✅ Valid local absolute file path
        ("file://C://One Drive//path//to//file", True),  # ✅ Valid local relative path
        ("file://./parent/directory/script.sh", False),  # ✅ relative path
    ],
)

def test_is_local_file_uri(uri, expected):
    assert(utilities.is_local_file_uri(uri)) == expected


# #
# # @pytest.mark.parametrize(
# #     "path, expected",
# #     [
# #         ("C:/Users/oca/NAPA-OCX_M1.3docx", True) if os.name == 'nt' else ("home/user/file.txt", False),
# #         ("D:/Users/oca/NAPA-OCX_M1.3docx", True)if os.name == 'nt' else ("./relative/path/to/file", True),
# #         ("C:Windows/System32/", False) if os.name == 'nt' else ("/var/log/system.log", True),
# #     ],
# # )
#
# def test_is_valid_file_path(path, expected):
#     assert(utilities.is_valid_file_path(path)) == expected

def test_file_uri_to_path_1():
    assert utilities.file_uri_to_path("file:C:/User/john/myfile.txt") == "C:\\User\\john\\myfile.txt"

def test_file_uri_to_path_2():
    assert utilities.file_uri_to_path("file:///C:/User/john/myfile.txt") == "C:\\User\\john\\myfile.txt"

def test_file_uri_to_path_3():
    assert utilities.file_uri_to_path("file://server/share/myfile.txt") == "server\\share\\myfile.txt"

def test_file_uri_to_path_4():
    assert utilities.file_uri_to_path("file:///home/user/myfile.txt") == "\\home\\user\\myfile.txt"
